#include "Arcball.h"

namespace moglclass
{
	Arcball::Arcball(int w, int h)
	{
		isRotating = false;
		setBounds(w, h);
		reset();
	}

	Arcball::~Arcball()
	{
	}
	
	// set the arcball window width x height
	void Arcball::setBounds(int w, int h)
	{
		width = w;
		height = h;
		ballRadius = w < h ? (w / 2) : (h / 2);
	}

	// reset the transformation matrix to the identity
	void Arcball::reset()
	{
		currentMatrix = glm::mat4(1.0f);
	}

	// get the rotation matrix
	glm::mat4 Arcball::getRotationMatrix()
	{
		if (isRotating)
		{
			if (glm::length(currentRotationVector - startRotationVector) > 1E-6)
			{
				// rotation axis
				glm::vec3 rotationAxis = glm::cross(currentRotationVector, startRotationVector);
				rotationAxis = glm::normalize(rotationAxis);
				// rotation angle
				float val = glm::dot(currentRotationVector, startRotationVector);
				val > (1 - 1E-10) ? val = 1.0 : val = val;
				float rotationAngle = acosf(val) * 180.0f / (float)M_PI;
				currentMatrix = glm::rotate(glm::mat4(1.0f), rotationAngle * 2, rotationAxis);
			}
		}
		return currentMatrix;
	}

	// map a mouse point (mx,my) to a point on a sphere (x, y, z)
	// if the points lie outside the sphere, are the window centric coordinates of the mouse
	glm::vec3 Arcball::convertXY(int mx, int my)
	{
		// 0~width(x+ right) x 0~height(y+ down) => -width/2~width/2(x+ right) x -height/2~height/2(y+ up)
		int w = glutGet(GLUT_WINDOW_WIDTH);
		int h = glutGet(GLUT_WINDOW_HEIGHT);
		int x = mx - (w / 2);
		int y = (h / 2) - my;

		// get a point on a sphere
		float d = (float)(x*x + y*y);
		glm::vec3 rotVector = glm::vec3(x, y, 0);
		float radiusSquared = ballRadius*ballRadius;
		if (d <= radiusSquared)
		{
			rotVector = glm::vec3((float)x, (float)y, sqrt(radiusSquared - d));
		}
		rotVector = glm::normalize(rotVector);

		return rotVector;
	}

	// when mouse click to startRotation
	void Arcball::startRotation(int mx, int my)
	{
		startRotationVector = convertXY(mx, my);
		currentRotationVector = startRotationVector;
		isRotating = true;
	}

	// when mouse drag to updateRotation
	void Arcball::updateRotation(int mx, int my)
	{
		currentRotationVector = convertXY(mx, my);
	}

	// when mouse release to stopRotation
	void Arcball::stopRotation()
	{
		isRotating = false;
	}
}